"""
Neural Bridge
=============

Submodule: functions_

"""

from collections.abc import Iterable
from os import remove as os_remove
from shutil import rmtree

from onnx import load as onnx_load, save as onnx_save
from onnx.onnx_ml_pb2 import ModelProto as Model_onnx
from onnx_tf.backend import prepare as convert_onnx_to_tf # pip install tensorflow-probability
from onnx2torch import convert as convert_onnx_to_torch
from tensorflow import Module as Model_tf, saved_model as tf_saved_model
from tf2onnx.convert import from_keras as convert_tf_to_onnx
from torch import dtype as torch_dtype, float32 as torch_float32
from torch import randn as torch_randn
from torch.jit import ScriptFunction as torch_ScriptFunction
from torch.jit import ScriptModule as torch_ScriptModule
from torch.nn import Module as Model_torch
from torch.onnx import export as torch_onnx_export

from .constants import TF_MODEL, TORCH_MODEL
from .constants import SupportedModel


def get_model_type_(model: SupportedModel) -> str:
    """Identify the type of the given model

    Parameters
    ----------
    model : SupportedModel
        Model to identify type of

    Returns
    -------
    str
        The identified model type

    Raises
    ------
    RuntimeError
        When type of the model cannot get identified.
    """

    result_ = TF_MODEL if isinstance(model, Model_tf) else\
              TORCH_MODEL if isinstance(model, Model_torch) else None
    if result_ is None:
        raise RuntimeError('Cannot determine model type. It is unsupported.')
    return result_


def onnx_to_tf_(model: Model_onnx, delete_tmp_dir : bool = True) -> Model_tf:
    """Convert ONNX model to TensorFlow

    Parameters
    ----------
    model : Model_onnx
        ONNX model instance to convert
    delete_tmp_dir : bool, optional
        Whether to delete temporarily created graph directory or not,
        by default True

    Returns
    -------
    Model_tf
        The converted TensorFlow model instance

    Notes
    -----
    This solution is based on:
        https://github.com/onnx/onnx-tensorflow/?tab=readme-ov-file#convert-programmatically
    """

    TMP_DIRECTORY_NAME_ = 'model_tmp_dir'
    convert_onnx_to_tf(model).export_graph(TMP_DIRECTORY_NAME_)
    result_ = tf_saved_model.load(TMP_DIRECTORY_NAME_)
    if delete_tmp_dir:
        __remove_directory(TMP_DIRECTORY_NAME_)
    return result_


def onnx_to_torch_(model: Model_onnx) -> Model_torch:
    """Convert ONNX model to Pytorch

    Parameters
    ----------
    model : Model_onnx
        ONNX model instance to convert

    Returns
    -------
    Model_torch
        The converted PyTorch model instance

    Notes
    -----
    This solution is based on:
        https://github.com/ENOT-AutoDL/onnx2torch
    """

    return convert_onnx_to_torch(model)


def tf_to_onnx_(model: Model_tf, onnx_path_base: str | None) -> Model_onnx:
    """Convert TensorFlow model to ONNX

    Parameters
    ----------
    model : Model_tf
        TensorFlow model instance to convert
    onnx_path_base : str | None
        Path to save the converted ONNX model instance to, or None. If set to
        None, ONNX model is not saved

    Returns
    -------
    Model_onnx
        The converted ONNX model instance

    Notes
    -----
    This solution is based on:
        https://github.com/onnx/tensorflow-onnx?tab=readme-ov-file#python-api-reference
    """

    result_, _ = convert_tf_to_onnx(model)
    if onnx_path_base is not None:
        onnx_save(result_, f'{onnx_path_base}.onnx')
    return result_


def torch_to_onnx_(model: Model_torch, shape: Iterable | None,
                   dtype: torch_dtype | str | None,
                   onnx_path_base: str | None,
                   delete_tmp_file: bool = True) -> Model_onnx:
    """Convert PyTorch model to ONNX

    Parameters
    ----------
    model : Model_torch
        PyTorch model instance to convert
    shape : Iterable | None
        Input shape for the given model, must be set only if given model is not
        a ScriptFunction or ScriptModel
    dtype : torch_dtype | str | None
        Input data type for given model, used only if given model is not a
        ScriptFunction or ScriptModel, if set to None torch.float32 is used
    onnx_path_base : str | None
        Path to save the converted ONNX model instance to, or None. If set to
        None, ONNX model is not saved
    delete_tmp_file : bool, optional
        Whether to delete temporarily created ONNX file or not, by default True

    Returns
    -------
    Model_onnx
        The converted ONNX model instance

    Raises
    ------
    RuntimeError
        When the given model is not a ScriptFunction or ScriptModel and shape is
        set to None.

    Notes
    -----
        https://pytorch.org/docs/stable/onnx_torchscript.html#torch.onnx.export
    """

    TMP_FILE_NAME_ = 'model.tmp'
    args_ = ()
    if shape is None:
        if not isinstance(model, (torch_ScriptFunction, torch_ScriptModule)):
            raise RuntimeError('PyTorch models require input shape if they are not script models.')
    else:
        dtype_ = torch_float32 if dtype is None else dtype
        args_ = torch_randn(shape, dtype=dtype_, requires_grad=True)
    torch_onnx_export(model, args_, TMP_FILE_NAME_, export_params=True,
                      opset_version=10, do_constant_folding=True,
                      input_names = ['input'], output_names = ['output'],
                      dynamic_axes={'input' : {0 : 'batch_size'},
                                    'output' : {0 : 'batch_size'}})
    result_ = onnx_load(TMP_FILE_NAME_)
    if delete_tmp_file:
        __remove_file(TMP_FILE_NAME_)
    if onnx_path_base is not None:
        onnx_save(result_, f'{onnx_path_base}.onnx')
    return result_


def __remove_directory(directory_path : str):
    """Remove (non-empty) directory

    Parameters
    ----------
    directory_path : str
        Name or path of the directory to remove
    """
    
    rmtree(directory_path, True)


def __remove_file(file_path : str) -> bool:
    """Remove file

    Parameters
    ----------
    file_path : str
        Name or path of the file to remove

    Returns
    -------
    bool
        Whether the removal was successful or not
    """

    try:
        os_remove(file_path)
        result_ = True
    except OSError:
        result_ = False
    return result_


"""
Neural Bridge
=============

Function: constants

"""

from typing import Union

from onnx.onnx_ml_pb2 import ModelProto as Model_onnx
from tensorflow import Module as Model_tf, Tensor as Tensor_tf
from torch import Tensor as Tensor_torch
from torch.nn import Module as Model_torch


AUTO_MODEL_TYPE = 'auto'
ONNX_MODEL = 'onnx'
TF_MODEL = 'tf'
TORCH_MODEL = 'torch'

SupportedModel = Union[Model_tf, Model_torch, Model_onnx]
SupportedTensor = Union[Tensor_tf, Tensor_torch]


"""
Neural Bridge
=============

Function: convert_model
"""

from collections.abc import Iterable

from torch import dtype as torch_dtype

from .constants import AUTO_MODEL_TYPE, TF_MODEL, TORCH_MODEL
from .constants import SupportedModel
from .functions_ import get_model_type_, onnx_to_tf_, onnx_to_torch_
from .functions_ import tf_to_onnx_, torch_to_onnx_


def convert_model(model: SupportedModel, *,
                  input_model_type: str = AUTO_MODEL_TYPE,
                  input_model_weights: str | None = None,
                  output_model_type: str = AUTO_MODEL_TYPE,
                  output_name: str | None = None, output_path: str | None = None,
                  save_onnx_output: bool = False,
                  torch_shape: Iterable | None = None,
                  torch_type: torch_dtype | str | None = None) -> SupportedModel:
    """Convert model from a supported model type to another supported model type

    Parameters
    ----------
    model : SupportedModel
        Model to convert
    input_model_type : str, optional
        Expected type to the given model, by default AUTO_MODEL_TYPE
    input_model_weights : str | None, optional
        Path to model weights, if None no weights are loaded, by default None
    output_model_type : str, optional
        Expected output model type, by default AUTO_MODEL_TYPE
    output_name : str | None, optional
        Name of the output model file(s) if set to None, class name will be
        used, by default None
    output_path : str | None, optional
        Path of model files to save to, if set to None, current directory will
        be used, by default None
    save_onnx_output : bool, optional
        Whether to save ONNX model or not, by default False
    torch_shape : Iterable | None, optional
        Input shape for PyTorch model, must be set only if given model is a
        PyTorch model which is not yet converted to ScriptFunction or
        ScriptModel, by default None
    torch_type : torch_dtype | str | None, optional
        Input data type for PyTorch model, must be set only if given model is a
        PyTorch model which is not yet converted to ScriptFunction or
        ScriptModel, by default None

    Returns
    -------
    SupportedModel
        The converted model

    Raises
    ------
    RuntimeError
        When expected given model type and given model type is different. This
        is a way to check real given model type. To avoid this error, leave
        input_model_type parameter to AUTO_MODEL_TYPE (default).
    ValueError
        When output model type is not supported.
    RuntimeError
        When conversion of the given model is not supported because if its type.
    RuntimeError
        When conversion of the output model is not supported because if its type.
    RuntimeError
        When conversion to ONNX model is not supported for the given model based
        on its type.

    Notes
    -----
    To see all the possible errors, read the documentation of the following
    functions:
        .functions_.get_model_type_()
        __determine_output_model_type()
        __load_model_weights()
        .functions_.torch_to_onnx_()
    """

    expected_type_ = __get_expected_input_type(input_model_type)
    input_model_type_ = get_model_type_(model)
    if expected_type_ != AUTO_MODEL_TYPE\
       and input_model_type_ != input_model_type:
        raise RuntimeError(f'Inconsistent types: expected input "{expected_type_}", given input "{input_model_type_}".')
    if output_model_type not in [AUTO_MODEL_TYPE, TF_MODEL, TORCH_MODEL]:
        raise ValueError(f'Unsupported output model type "{output_model_type}".')
    output_model_type_ = __determine_output_model_type(input_model_type_,
                                                       output_model_type)
    output_path_base_ = __get_output_path_base(model, output_name, output_path)
    onnx_path_base_ = output_path_base_ if save_onnx_output else None
    if input_model_weights is not None:
        __load_model_weights(model, input_model_type_, input_model_weights)
    if input_model_type_ != output_model_type_:
        if input_model_type_ == TF_MODEL:
            onnx_model_ = tf_to_onnx_(model, onnx_path_base_)
        elif input_model_type_ == TORCH_MODEL:
            onnx_model_ = torch_to_onnx_(model, torch_shape, torch_type,
                                         onnx_path_base_)
        else:
            raise RuntimeError(f'Conversion for model type "{input_model_type_}" is not supported.')
        if output_model_type_ == TF_MODEL:
            result_ = onnx_to_tf_(onnx_model_)
        elif output_model_type_ == TORCH_MODEL:
            result_ = onnx_to_torch_(onnx_model_)
        else:
            raise RuntimeError(f'Conversion for model type "{input_model_type_}" is not supported.')
    else:
        if save_onnx_output == True:
            if input_model_type_ == TF_MODEL:
                tf_to_onnx_(model, onnx_path_base_)
            elif input_model_type_ == TORCH_MODEL:
                torch_to_onnx_(model, torch_shape, torch_type, onnx_path_base_)
            else:
                raise RuntimeError(f'ONNX export for model type "{input_model_type_}" is not supported.')
        result_ = model
    return result_


def __determine_output_model_type(given_model_type: str,
                                  output_model_type: str) -> str:
    """Determine expected output model type based on given model type and output
    model type parameter

    Parameters
    ----------
    given_model_type : str
        Type of the given model
    output_model_type : str
        The given output type parameter

    Returns
    -------
    str
        Model type that the convert_model() function expects to output

    Raises
    ------
    ValueError
        When the expected output model type cannot get determined. If output
        model type is set to AUTO_MODEL_TYPE, expected output is determined by
        the type of the given model.
    """

    result_ = output_model_type
    if result_ == AUTO_MODEL_TYPE:
        if given_model_type not in [TF_MODEL, TORCH_MODEL]:
            raise ValueError(f'Cannot determine desired output model type because given model is "{given_model_type}".')
        else:
            result_ = TF_MODEL if given_model_type == TORCH_MODEL\
                               else TORCH_MODEL
    return result_


def __get_expected_input_type(input_model_type: str) -> str:
    """Determine the expected model type

    Parameters
    ----------
    input_model_type : str
        Input type parameter set by the user

    Returns
    -------
    str
        Converted model type as input expectation
    """

    return input_model_type if input_model_type in [AUTO_MODEL_TYPE, TF_MODEL,
                                                    TORCH_MODEL]\
                            else AUTO_MODEL_TYPE


def __get_output_path_base(model: SupportedModel, output_name: str | None,
                           output_path: str | None) -> str:
    """_summary_

    Parameters
    ----------
    model : SupportedModel
        The given model
    output_name : str | None
        Output model name, if set to None, class name will be used
    output_path : str | None
        Output path, if set to None, current path will be used

    Returns
    -------
    str
        Path base to use for saving the model
    """

    output_name_ = output_name if output_name is not None\
                               else model.__class__.__name__
    output_path_ = output_path if output_path is not None else './'
    return f'{output_path_.rstrip("/")}/{output_name_}'


def __load_model_weights(model: SupportedModel, model_type: str,
                         path_to_weights: str):
    """Load model weights

    Parameters
    ----------
    model : SupportedModel
        Model to load weights for
    model_type : str
        Type of the model to load weights for
    path_to_weights : str
        Path of the file of the model weights

    Raises
    ------
    RuntimeError
        When loading weights is not supported for the given model.
    """

    if model_type == TF_MODEL:
        model.load_weights(path_to_weights)
    elif model_type == TORCH_MODEL:
        model.load_state_dict(path_to_weights)
    else:
        raise RuntimeError(f'Loading separated weights for model type "{model_type}" is not supported.')


"""
Neural Bridge
=============

Function: constants
"""

from onnx.onnx_ml_pb2 import ModelProto as Model_onnx
from tensorflow import Module as Model_tf
from torch.nn import Module as Model_torch

from .constants import ONNX_MODEL, TF_MODEL, TORCH_MODEL
from .constants import SupportedModel
from .functions_ import get_model_type_, onnx_to_tf_, onnx_to_torch_
from .functions_ import tf_to_onnx_, torch_to_onnx_


def load_as(model: SupportedModel, output_type: str) -> SupportedModel:
    """Load model as the specified type

    Parameters
    ----------
    model : SupportedModel
        Model to load
    output_type : str
        Type of the model to load as

    Returns
    -------
    SupportedModel
        The loaded model with type according to the output_type parameter

    Raises
    ------
    RuntimeError
        When unsupported model type is given.
    """

    if output_type == ONNX_MODEL:
        return load_as_onnx(model)
    elif output_type == TF_MODEL:
        return load_as_tf(model)
    elif output_type == TORCH_MODEL:
        return load_as_torch(model)
    else:
        raise RuntimeError(f'Unsupported output type "{output_type}".')


def load_as_onnx(model: SupportedModel) -> Model_onnx:
    """Load model as ONNX Model

    Parameters
    ----------
    model : SupportedModel
        Model to load

    Returns
    -------
    Model_onnx
        The loaded ONNX model
    """

    model_type_ = get_model_type_(model)
    if model_type_ == TF_MODEL:
        return tf_to_onnx_(model, None)
    elif model_type_ == TORCH_MODEL:
        return torch_to_onnx_(model, None, None, None)
    elif model_type_ == ONNX_MODEL:
        return model


def load_as_tf(model: SupportedModel) -> Model_tf:
    """Load model as TensorFlow Model

    Parameters
    ----------
    model : SupportedModel
        Model to load

    Returns
    -------
    Model_tf
        The loaded TensorFlow model
    """

    model_type_ = get_model_type_(model)
    if model_type_ == TORCH_MODEL:
        onnx_model_ = torch_to_onnx_(model, None, None, None)
        return onnx_to_tf_(onnx_model_)
    elif model_type_ == ONNX_MODEL:
        return onnx_to_tf_(model)
    elif model_type_ == TF_MODEL:
        return model


def load_as_torch(model: SupportedModel) -> Model_torch:
    """PyTorch

    Parameters
    ----------
    model : SupportedModel
        Model to load

    Returns
    -------
    Model_torch
        The loaded PyTorch model
    """

    model_type_ = get_model_type_(model) 
    if model_type_ == TF_MODEL:
        onnx_model_ = tf_to_onnx_(model, None)
        return onnx_to_torch_(onnx_model_)
    elif model_type_ == ONNX_MODEL:
        return tf_to_onnx_(model, None)
    elif model_type_ == TORCH_MODEL:
        return model


"""
Neural Bridge
=============

Module main file.
"""

from .convert_model import convert_model
from .load import load_as, load_as_onnx, load_as_tf, load_as_torch


__all__ = [convert_model, load_as, load_as_onnx, load_as_tf, load_as_torch]